from .types import (GetAvailabilityInfoResponse, GetTitleDetailsResponse, Item,
                    SearchResponse, Title, TitleDetail)


def search_response_factory(response):
    try:
        titles = []
        for title in response['Titles']['Title']:
            titles.append(Title(
                bid=title['BID'],
                isbn=title['ISBN'],
                title_name=title['TitleName'],
                author=title['Author'],
                publish_year=title['PublishYear'],
                media_code=title['MediaCode'],
                media_desc=title['MediaDesc']
            ))
    except TypeError:
        titles = None
    return SearchResponse(
        status=response['Status'],
        message=response['Message'],
        error_message=response['ErrorMessage'],
        total_records=response['TotalRecords'],
        next_record_position=response['NextRecordPosition'],
        set_id=response['SetId'],
        titles=titles and tuple(titles)
    )


def get_title_details_response_factory(response):
    try:
        title_detail = TitleDetail(
            bid=response['TitleDetail']['BID'],
            title_name=response['TitleDetail']['TitleName'],
            author=response['TitleDetail']['Author'],
            other_authors=response['TitleDetail']['OtherAuthors'],
            publisher=response['TitleDetail']['Publisher'],
            physical_desc=response['TitleDetail']['PhysicalDesc'],
            subjects=tuple(response['TitleDetail']['Subjects']['Subject']),
            summary=response['TitleDetail']['Summary'],
            notes=response['TitleDetail']['Notes'],
            isbn=response['TitleDetail']['ISBN'],
            issn=response['TitleDetail']['ISSN'],
            n_title_name=response['TitleDetail']['NTitleName'],
            n_author=response['TitleDetail']['NAuthor'],
            n_publisher=response['TitleDetail']['NPublisher'],
        )
    except TypeError:
        title_detail = None
    return GetTitleDetailsResponse(
        status=response['Status'],
        message=response['Message'],
        error_message=response['ErrorMessage'],
        title_detail=title_detail
    )


def get_availability_info_response_factory(response):
    try:
        items = []
        for item in response['Items']['Item']:
            items.append(Item(
                item_no=item['ItemNo'],
                branch_id=item['BranchID'],
                branch_name=item['BranchName'],
                location_code=item['LocationCode'],
                location_desc=item['LocationDesc'],
                call_number=item['CallNumber'],
                status_code=item['StatusCode'],
                status_desc=item['StatusDesc'],
                media_code=item['MediaCode'],
                media_desc=item['MediaDesc'],
                status_date=item['StatusDate'],
                due_date=item['DueDate'],
                cluster_name=item['ClusterName'],
                category_name=item['CategoryName'],
                collection_code=item['CollectionCode'],
                collection_min_age_limit=item['CollectionMinAgeLimit'],
            ))
    except TypeError:
        items = None
    return GetAvailabilityInfoResponse(
        status=response['Status'],
        message=response['Message'],
        error_message=response['ErrorMessage'],
        next_record_position=response['NextRecordPosition'],
        set_id=response['SetId'],
        items=items and tuple(items),
    )



from typing import Union

from zeep import Client as Zeep

from ._factories import (get_availability_info_response_factory,
                         get_title_details_response_factory,
                         search_response_factory)
from .constants import Branch, Language, MediaCode, Sort
from .types import (GetAvailabilityInfoResponse, GetTitleDetailsResponse,
                    SearchResponse)

PRODUCTION_URL = 'https://openweb.nlb.gov.sg/OWS/CatalogueService.svc?singleWsdl'
STAGING_URL = 'https://openweb-stg.nlb.gov.sg/OWS/CatalogueService.svc?singleWsdl'


class Client:
    def __init__(self, wsdl: str, api_key: str) -> None:
        self._zeep = Zeep(wsdl)
        self.api_key = api_key

    @staticmethod
    def _search_request(api_key, keywords=None, author=None, subject=None, title=None,
                        branch=None, media_code=None, language=None,
                        sort=None, start=1, limit=10, set_id=None):
        search_items = []
        if keywords:
            search_items.append({
                'SearchField': 'Keywords',
                'SearchTerms': keywords,
            })
        if author:
            search_items.append({
                'SearchField': 'Author',
                'SearchTerms': author,
            })
        if subject:
            search_items.append({
                'SearchField': 'Subject',
                'SearchTerms': subject
            })
        if title:
            search_items.append({
                'SearchField': 'Title',
                'SearchTerms': title
            })
        if branch:
            search_items.append({
                'SearchField': 'BranchID',
                'SearchTerms': branch.value if isinstance(branch, Branch) else branch,
            })
        if media_code:
            search_items.append({
                'SearchField': 'MediaCode',
                'SearchTerms': media_code.value if isinstance(media_code, MediaCode) else media_code,
            })
        if language:
            search_items.append({
                'SearchField': 'Language',
                'SearchTerms': language.value if isinstance(language, Language) else language,
            })
        return {
            'APIKey': api_key,
            'SearchItems': {
                'SearchItem': search_items,
            },
            'Modifiers': {
                'SortSchema': sort.value if isinstance(sort, Sort) else sort,
                'StartRecordPosition': start,
                'MaximumRecords': limit,
                'SetId': set_id,
            }
        }

    def search(self, keywords: str = None, author: str = None, subject: str = None, title: str = None,
               branch: Union[str, Branch] = None, media_code: Union[str, MediaCode] = None,
               language: Union[str, Language] = None,
               sort: Union[str, Sort] = None, start: int = 1, limit: int = 10, set_id: str = None) -> SearchResponse:
        """
        Searches content according to search criteria.

        :param keywords:
        :param author:
        :param subject:
        :param title:
        :param branch: Include only results from a particular branch. See `constants.Branch` for possible values.
        :param media_code: Include only results of a particular media type. See `constants.MediaCode` for possible
                           values.
        :param language: Include only results in a particular language. See `constants.Language` for possible values.
        :param sort: By default, results are sorted by published year in descending order. Use ``PUBDATE`` to sort in
                     ascending order instead, or ``TITLE`` to sort by title in ascending order, ignoring articles like
                     "a", "an" or "the". These values can also be found in `constants.Sort`.
        :param start: Start pointer for returned records.
        :param limit: Maximum records to be returned. This is capped at 100 records even if a number greater than 100
                      is specified.
        :param set_id: For use in pagination. This can be used with ``start`` to return the index position of
                       the next record in the backend system
        :return:
        """
        search_request = self._search_request(self.api_key, keywords, author, subject, title,
                                              branch, media_code, language,
                                              sort, start, limit, set_id)
        response = self._zeep.service.Search(**search_request)
        return search_response_factory(response)

    def get_title_details(self, bid: str = None, isbn: str = None) -> GetTitleDetailsResponse:
        """
        Get detailed information about an item. Either ``bid`` or ``isbn`` should be provided.

        :param bid:
        :param isbn:
        :return:
        """
        response = self._zeep.service.GetTitleDetails(APIKey=self.api_key,
                                                      BID=bid,
                                                      ISBN=isbn)
        return get_title_details_response_factory(response)

    def get_availability_info(self, bid: str = None, isbn: str = None,
                              sort: Union[str, Sort] = None, start: int = 1, limit: int = 10,
                              set_id: str = None) -> GetAvailabilityInfoResponse:
        """
        Check whether an item is available for loan. Either ``bid`` or ``isbn`` should be provided.

        :param bid:
        :param isbn:
        :param sort: By default, results are sorted by published year in descending order. Use ``PUBDATE`` to sort in
                     ascending order instead, or ``TITLE`` to sort by title in ascending order, ignoring articles like
                     "a", "an" or "the". These values can also be found in `constants.Sort`.
        :param start: Start pointer for returned records.
        :param limit: Maximum records to be returned. This is capped at 100 records even if a number greater than 100
                      is specified.
        :param set_id: For use in pagination. This can be used with ``start`` to return the index position of
                       the next record in the backend system
        :return:
        """
        response = self._zeep.service.GetAvailabilityInfo(APIKey=self.api_key,
                                                          BID=bid,
                                                          ISBN=isbn,
                                                          Modifiers={
                                                              'SortSchema': sort.value if isinstance(sort,
                                                                                                     Sort) else sort,
                                                              'StartRecordPosition': start,
                                                              'MaximumRecords': limit,
                                                              'SetId': set_id,
                                                          })
        return get_availability_info_response_factory(response)



from dataclasses import dataclass
from typing import Iterable, Optional


@dataclass
class Title:
    """
Part of `SearchResponse`.

:var str bid:
:var str isbn:
:var str title_name:
:var str author:
:var str publish_year:
:var str media_code:
:var str media_desc:

Example `Title`::

    Title(
        bid='203125808',
        isbn='1328915336 (paperback)',
        title_name='Beren and Lúthien / by  J.R.R. Tolkien ; edited by Christopher Tolkien ; with illustrations by  Alan Lee.',
        author='Tolkien, J. R. R. (John Ronald Reuel), 1892-1973',
        publish_year='2018',
        media_code='BK',
        media_desc='Books'
    )

    """
    bid: str
    isbn: str
    title_name: str
    author: str
    publish_year: str
    media_code: str
    media_desc: str


@dataclass
class SearchResponse:
    """
Returned by `Client.search`.

:var str status:
:var str message:
:var Optional[str] error_message:
:var Optional[int] total_records:
:var Optional[int] next_record_position:
:var Optional[str] set_id:
:var Optional[Iterable[Title]] titles:

Example `SearchResponse`::

    SearchResponse(
        status='OK',
        message='Operation completed successfully',
        error_message=None,
        total_records=52,
        next_record_position=4,
        set_id='PGE3676',
        titles=tuple([
            Title(
                bid='203125808',
                isbn='1328915336 (paperback)',
                title_name='Beren and Lúthien / by  J.R.R. Tolkien ; edited by Christopher Tolkien ; with illustrations by  Alan Lee.',
                author='Tolkien, J. R. R. (John Ronald Reuel), 1892-1973',
                publish_year='2018',
                media_code='BK',
                media_desc='Books'
            ),
            Title(
                bid='204576140',
                isbn='9780008214210 (electronic bk)',
                title_name='Beren and l℗♭©ʻthien [electronic resource]. J. R. R Tolkien.',
                author='Tolkien, J. R. R.',
                publish_year='2017',
                media_code='BK',
                media_desc='Books'
            )
        ])
    )

    """
    status: str
    message: str
    error_message: Optional[str]
    total_records: Optional[int]
    next_record_position: Optional[int]
    set_id: Optional[str]
    titles: Optional[Iterable[Title]]


@dataclass
class TitleDetail:
    """
Part of `GetTitleDetailsResponse`.

:var str bid:
:var str title_name:
:var str author:
:var str other_authors:
:var Optional[str] publisher:
:var str physical_desc:
:var Iterable[str] subjects:
:var str summary:
:var str notes:
:var str isbn:
:var Optional[str] issn:
:var Optional[str] n_title_name:
:var Optional[str] n_author:
:var Optional[str] n_publisher:

Example `TitleDetail`::

    TitleDetail(
        bid='203125808',
        title_name='Beren and Lúthien / by  J.R.R. Tolkien ; edited by Christopher Tolkien ; with illustrations by  Alan Lee.',
        author='Tolkien, J. R. R.',
        other_authors='Tolkien, J. R. R. (John Ronald Reuel), 1892-1973|Tolkien, Christopher|Lee, Alan',
        publisher=None,
        physical_desc='288 pages (pages numbered 8-288) :chiefly color illustrations ;21 cm',
        subjects=(
            'Middle Earth (Imaginary place) Fiction',
            'Elves Fiction',
            'Fantasy fiction'
        ),
        summary="The epic tale of Beren and Lúthien became an essential element in the evolution of The Silmarillion, the myths and legends of J.R.R. Tolkien's First Age of the World. Always key to the story is the fate that shadowed their love: Beren was a mortal man, Lúthien an immortal Elf. Her father, a great Elvish lord, imposed on Beren an impossible task before he might wed Lúthien: to rob the greatest of all evil beings, Melkor, of a Silmaril.Painstakingly restored from Tolkien's manuscripts and presented for the first time as a continuous and standalone story, Beren and Lúthien reunites fans of The Hobbit and The Lord of the Rings with Elves and Men, along with the rich landscape and creatures unique to Tolkien's Middle-earth. Christopher Tolkien tells the story in his father's own words by giving its original form as well as prose and verse passages from later texts that illustrate the narrative as it changed. -- from back cover.",
        notes="First published by Harper Collins Publishers 2017.Includes abstractsThe epic tale of Beren and Lúthien became an essential element in the evolution of The Silmarillion, the myths and legends of J.R.R. Tolkien's First Age of the World. Always key to the story is the fate that shadowed their love: Beren was a mortal man, Lúthien an immortal Elf. Her father, a great Elvish lord, imposed on Beren an impossible task before he might wed Lúthien: to rob the greatest of all evil beings, Melkor, of a Silmaril.Painstakingly restored from Tolkien's manuscripts and presented for the first time as a continuous and standalone story, Beren and Lúthien reunites fans of The Hobbit and The Lord of the Rings with Elves and Men, along with the rich landscape and creatures unique to Tolkien's Middle-earth. Christopher Tolkien tells the story in his father's own words by giving its original form as well as prose and verse passages from later texts that illustrate the narrative as it changed. -- from back cover.",
        isbn='1328915336 (paperback)',
        issn=None,
        n_title_name=None,
        n_author=None,
        n_publisher=None
    )

    """
    bid: str
    title_name: str
    author: str
    other_authors: str
    publisher: Optional[str]
    physical_desc: str
    subjects: Iterable[str]
    summary: str
    notes: str
    isbn: str
    issn: Optional[str]
    n_title_name: Optional[str]
    n_author: Optional[str]
    n_publisher: Optional[str]


@dataclass
class GetTitleDetailsResponse:
    """
Returned by `Client.get_title_details`.

:var str status:
:var str message:
:var Optional[str] error_message:
:var Optional[TitleDetail] title_detail:

Example `GetTitleDetailsResponse`::

    GetTitleDetailsResponse(
        status='OK',
        message='Operation completed successfully',
        error_message=None,
        title_detail=TitleDetail(
            bid='203125808',
            title_name='Beren and Lúthien / by  J.R.R. Tolkien ; edited by Christopher Tolkien ; with illustrations by  Alan Lee.',
            author='Tolkien, J. R. R.',
            other_authors='Tolkien, J. R. R. (John Ronald Reuel), 1892-1973|Tolkien, Christopher|Lee, Alan',
            publisher=None,
            physical_desc='288 pages (pages numbered 8-288) :chiefly color illustrations ;21 cm',
            subjects=(
                'Middle Earth (Imaginary place) Fiction',
                'Elves Fiction',
                'Fantasy fiction'
            ),
            summary="The epic tale of Beren and Lúthien became an essential element in the evolution of The Silmarillion, the myths and legends of J.R.R. Tolkien's First Age of the World. Always key to the story is the fate that shadowed their love: Beren was a mortal man, Lúthien an immortal Elf. Her father, a great Elvish lord, imposed on Beren an impossible task before he might wed Lúthien: to rob the greatest of all evil beings, Melkor, of a Silmaril.Painstakingly restored from Tolkien's manuscripts and presented for the first time as a continuous and standalone story, Beren and Lúthien reunites fans of The Hobbit and The Lord of the Rings with Elves and Men, along with the rich landscape and creatures unique to Tolkien's Middle-earth. Christopher Tolkien tells the story in his father's own words by giving its original form as well as prose and verse passages from later texts that illustrate the narrative as it changed. -- from back cover.",
            notes="First published by Harper Collins Publishers 2017.Includes abstractsThe epic tale of Beren and Lúthien became an essential element in the evolution of The Silmarillion, the myths and legends of J.R.R. Tolkien's First Age of the World. Always key to the story is the fate that shadowed their love: Beren was a mortal man, Lúthien an immortal Elf. Her father, a great Elvish lord, imposed on Beren an impossible task before he might wed Lúthien: to rob the greatest of all evil beings, Melkor, of a Silmaril.Painstakingly restored from Tolkien's manuscripts and presented for the first time as a continuous and standalone story, Beren and Lúthien reunites fans of The Hobbit and The Lord of the Rings with Elves and Men, along with the rich landscape and creatures unique to Tolkien's Middle-earth. Christopher Tolkien tells the story in his father's own words by giving its original form as well as prose and verse passages from later texts that illustrate the narrative as it changed. -- from back cover.",
            isbn='1328915336 (paperback)',
            issn=None,
            n_title_name=None,
            n_author=None,
            n_publisher=None
        )
    )

    """
    status: str
    message: str
    error_message: Optional[str]
    title_detail: Optional[TitleDetail]


@dataclass
class Item:
    """
Part of `GetAvailabilityInfoResponse`.

:var str item_no:
:var str branch_id:
:var str branch_name:
:var str location_code:
:var str location_desc:
:var str call_number:
:var str status_code:
:var str status_desc:
:var Optional[str] media_code:
:var str media_desc:
:var str status_date:
:var Optional[str] due_date:
:var Optional[str] cluster_name:
:var Optional[str] category_name:
:var Optional[str] collection_code:
:var Optional[str] collection_min_age_limit:

Example `Item`::

    Item(
        item_no='B33315118C',
        branch_id='BBPL',
        branch_name='Bukit Batok Public Library',
        location_code='____',
        location_desc='Adult Lending',
        call_number='English      TOL -[FN]',
        status_code='C',
        status_desc='On Loan',
        media_code=None,
        media_desc='Book',
        status_date='08/11/2018',
        due_date='08/11/2018',
        cluster_name=None,
        category_name=None,
        collection_code=None,
        collection_min_age_limit=None
    )

    """
    item_no: str
    branch_id: str
    branch_name: str
    location_code: str
    location_desc: str
    call_number: str
    status_code: str
    status_desc: str
    media_code: Optional[str]
    media_desc: str
    status_date: str
    due_date: Optional[str]
    cluster_name: Optional[str]
    category_name: Optional[str]
    collection_code: Optional[str]
    collection_min_age_limit: Optional[str]


@dataclass
class GetAvailabilityInfoResponse:
    """
Returned by `Client.get_availability_info`.

:var str status:
:var str message:
:var Optional[str] error_message:
:var Optional[int] next_record_position:
:var Optional[str] set_id:
:var Optional[Iterable[Item]] items:

Example `GetAvailabilityInfoResponse`::

    GetAvailabilityInfoResponse(
        status='OK',
        message='Operation completed successfully',
        error_message=None,
        next_record_position=2,
        set_id='3709',
        items=tuple([
            Item(
                item_no='B33315114J',
                branch_id='AMKPL',
                branch_name='Ang Mo Kio Public Library',
                location_code='____',
                location_desc='Adult Lending',
                call_number='English      TOL -[FN]',
                status_code='S',
                status_desc='Not On Loan',
                media_code=None,
                media_desc='Book',
                status_date='06/09/2018',
                due_date=None,
                cluster_name=None,
                category_name=None,
                collection_code=None,
                collection_min_age_limit=None
            ),
            Item(
                item_no='B33315118C',
                branch_id='BBPL',
                branch_name='Bukit Batok Public Library',
                location_code='____',
                location_desc='Adult Lending',
                call_number='English      TOL -[FN]',
                status_code='C',
                status_desc='On Loan',
                media_code=None,
                media_desc='Book',
                status_date='08/11/2018',
                due_date='08/11/2018',
                cluster_name=None,
                category_name=None,
                collection_code=None,
                collection_min_age_limit=None
            )
        ])
    )

    """
    status: str
    message: str
    error_message: Optional[str]
    next_record_position: Optional[int]
    set_id: Optional[str]
    items: Optional[Iterable[Item]]



from setuptools import setup

with open('docs/index.rst') as f:
    s = f.read()
    long_description = s.split('\n.. LONG DESCRIPTION ENDS HERE')[0]

version = {}
with open("nlbsg/__version__.py") as f:
    exec(f.read(), version)

setup(name='nlbsg',
      version=version['__version__'],
      description='Python SDK for the NLB Open Web Services',
      long_description_content_type='text/x-rst',
      long_description=long_description,
      url='https://github.com/yi-jiayu/nlbsg',
      author='Jiayu Yi',
      author_email='yijiayu@gmail.com',
      license='MIT',
      classifiers=[
          "Programming Language :: Python :: 3.7",
          "License :: OSI Approved :: MIT License",
          "Operating System :: OS Independent",
      ],
      project_urls={
          'Documentation': 'https://nlbsg.readthedocs.io/en/latest/',
          'Source code': 'https://github.com/yi-jiayu/nlbsg',
          'Builds': 'https://travis-ci.com/yi-jiayu/nlbsg',
          'Coverage': 'https://codecov.io/gh/yi-jiayu/nlbsg',
      },
      packages=['nlbsg'],
      install_requires=[
          'zeep',
      ],
      python_requires=">=3.7",
      zip_safe=False)



from enum import Enum


class Language(Enum):
    ENGLISH = 'English'
    CHINESE = 'Chinese'
    MALAY = 'Malay'
    TAMIL = 'Tamil'


class MediaCode(Enum):
    """
    Title Level Media Code and Media Description Mapping

    http://www.nlb.gov.sg/labs/technical-documentation/#TitleLevelMediaCodeandMediaDescriptionMapping
    """
    BOOKS = 'BK'
    COMPUTER_FILE = 'CF'
    MAPS = 'MP'
    MUSIC = 'MU'
    MIXED_MATERIALS = 'MX'
    SERIALS = 'SE'
    VISUAL_MATERIALS = 'VM'


class Branch(Enum):
    """
    Branch Code and Branch Name Mapping

    http://www.nlb.gov.sg/labs/technical-documentation/#BranchCodeandBranchNameMapping
    """
    TAMPINES_REGIONAL_LIBRARY = 'TRL'
    WOODLANDS_REGIONAL_LIBRARY = 'WRL'
    CLEMENTI_PUBLIC_LIBRARY = 'CMPL'
    LEE_KONG_CHIAN_REFERENCE_LIBRARY = 'LKCRL'
    ANG_MO_KIO_PUBLIC_LIBRARY = 'AMKPL'
    BUKIT_BATOK_PUBLIC_LIBRARY = 'BBPL'
    BUKIT_PANJANG_PUBLIC_LIBRARY = 'BPPL'
    BUKIT_MERAH_PUBLIC_LIBRARY = 'BMPL'
    CENTRAL_PUBLIC_LIBRARY = 'CLL'
    CHENG_SAN_PUBLIC_LIBRARY = 'CSPL'
    CHOA_CHU_KANG_PUBLIC_LIBRARY = 'CCKPL'
    GEYLANG_EAST_PUBLIC_LIBRARY = 'GEPL'
    JURONG_REGIONAL_LIBRARY = 'JRL'
    JURONG_WEST_PUBLIC_LIBRARY = 'JWPL'
    QUEENSTOWN_PUBLIC_LIBRARY = 'QUPL'
    TOA_PAYOH_PUBLIC_LIBRARY = 'TPPL'
    YISHUN_PUBLIC_LIBRARY = 'YIPL'
    LIBRARY_AT_ORCHARD = 'OCPL'
    LIBRARY_AT_CHINATOWN = 'CNPL'
    SEMBAWANG_PUBLIC_LIBRARY = 'SBPL'
    LIBRARY_SUPPLY_CENTRE = 'LSC'
    SERANGOON_PUBLIC_LIBRARY = 'SRPL'
    NL_HERITAGE = 'LOLC'
    MARINE_PARADE_PUBLIC_LIBRARY = 'MPPL'
    BEDOK_PUBLIC_LIBRARY = 'BEPL'
    SENGKANG_PUBLIC_LIBRARY = 'SKPL'
    LIBRARY_AT_ESPLANADE = 'EPPL'
    MOBILE_BUS = 'MOLLEY'
    PASIR_RIS_PUBLIC_LIBRARY = 'PRPL'
    BISHAN_PUBLIC_LIBRARY = 'BIPL'
    LIBRARY_SUPPLY_CENTRE_FOR_AV = 'LSCAV'
    LEE_KONG_CHIAN_REFERENCE_LIBRARY_LEVEL_7 = '07LKCRL'
    LEE_KONG_CHIAN_REFERENCE_LIBRARY_LEVEL_8 = '08LKCRL'
    LEE_KONG_CHIAN_REFERENCE_LIBRARY_LEVEL_9 = '09LKCRL'
    LEE_KONG_CHIAN_REFERENCE_LIBRARY_LEVEL_11 = '11LKCRL'


class Sort(Enum):
    PUBDATE = 'PUBDATE'
    TITLE = 'TITLE'
